import numpy as np
from numba import njit, prange
from numba import boolean, float64, float32, int32
from numba.experimental import jitclass
from numba.typed import List
from cupyx.scipy.ndimage import gaussian_filter
import cupy as cp
from numba.typed import List


# Class specification used for Numba's jitclass decorator
spec = [('x_center',float32),
        ('y_center',float32),
        ('z_center',float32),
        ('x',float64[:]),
        ('y',float64[:]),
        ('z',float64[:]),
        ('psi',float32),
        ('theta',float32),
        ('length',float32),
        ('radius',float32),
        ('internal_radius',float32)]

@jitclass(spec)
class CNT:
    '''A class to hold all values and parameters for a single MWCNT.
    Parameters
    ----------
    x_center : float32
        X-position of MWCNT center point.
    y_center : float32
        Y-position of MWCNT center point.
    z_center : float32
        Z-position of MWCNT center point.
    x : float64[:]
        Set of x-positions for points sampled along the MWCNT central axis.
    y : float64[:]
        Set of y-positions for points sampled along the MWCNT central axis.
    z : float64[:]
        Set of z-positions for points sampled along the MWCNT central axis.
    psi : float32
        Euler angle describing the MWCNT rotation from vertical into the XY-plane (about the Y-axis).
    theta : float32
        Euler angle describing the MWCNT rotation about the Z-axis.
    length : float32
        Length of the MWCNT.
    radius : float32
        Radius of the MWCNT.

    Returns
    -------
    Instance of the CNT class.
    '''
    def __init__(self, x_center, y_center, z_center, 
                 x, y, z, psi, theta, length, radius):
        self.x_center = x_center
        self.y_center = y_center
        self.z_center = z_center
        self.x = x
        self.y = y
        self.z = z
        self.psi = psi
        self.theta = theta
        self.length = length
        self.radius = radius


@njit(float64(float64, float64, float64, float64, float64, float64, int32, int32))
def calc_distance3D(x1, y1, z1, x2, y2, z2, boxsize, boxsize_z):
    '''Calculates 3D distance between two points, using periodic boundary conditions
    Parameters
    ----------
    x1 : float64
        x-position of point 1
    y1 : float64
        y-position of point 1
    z1 : float64
        z-position of point 1
    x2 : float64
        x-position of point 2
    y2 : float64
        y-position of point 2
    z2 : float64
        z-position of point 2
    boxsize : int32
        size of the periodic system in x and y
    boxsize_z : int32
        size of the periodic system in z

    Returns
    -------
    distance : float64
        The Euclidean distance between two points under periodic boundary conditions
    '''
    # periodic in x
    dx = abs(x1-x2)
    if dx > boxsize/2:
        min_x = min(x1,x2) + boxsize
        dx = min_x - max(x1,x2)
        
    # periodic in y
    dy = abs(y1-y2)
    if dy > boxsize/2:
        min_y = min(y1,y2) + boxsize
        dy = min_y - max(y1,y2)
        
    # periodic in z
    dz = abs(z1-z2)
    if dz > boxsize_z/2:
        min_z = min(z1,z2) + boxsize_z
        dz = min_z - max(z1,z2)
    return np.sqrt(dx**2 + dy**2 + dz**2)


@njit(boolean(float64[:], float64[:], float64[:], float64[:], float64[:], float64[:], int32, int32, int32))
def _intersect(x1, y1, z1, x2, y2, z2, min_dist, boxsize, boxsize_z):
    '''Determines if two sets of points are within a minimum distance of one another, using periodic boundary conditions
    Parameters
    ----------
    x1 : float64[:]
        X-positions for all points in first set.
    y1 : float64[:]
        Y-positions for all points in first set.
    z1 : float64[:]
        Z-positions for all points in first set.
    x2 : float64[:]
        X-positions for all points in second set.
    y2 : float64[:]
        Y-positions for all points in second set.
    z2 : float64[:]
        Z-positions for all points in second set.
    min_dist : int32
        Minimum distance below which two points are considered intersecting.
    boxsize : int32
        Size of periodic system in x and y.
    boxsize_z : int32
        Size of periodic system in z.

    Returns
    -------
    intersection: bool
        Returns True or False to whether the two sets of points are intersecting.
    '''
    for i in range(len(x1)):
        for j in range(len(x2)):
            test_dist = calc_distance3D(x1[i], y1[i], z1[i], x2[j], y2[j], z2[j], boxsize, boxsize_z)
            
            if test_dist < min_dist:
                return True
    return False


@njit()
def test_intersect(test_CNT, all_CNTs, boxsize, boxsize_z):
    '''Tests for intersection between a newly created CNT and all CNTs that have previously been placed into the box.
    Parameters
    ----------
    test_CNT : CNT
        New CNT which will be tested for intersection with all other CNTs.
    all_CNTs : List[CNT]
        List of CNTs that have previously been placed into the box.
    boxsize : int32
        Size of box in X and Y dimensions.
    boxsize_z : int32
        Size of box in Z dimension.

    Returns
    -------
    intersect : bool
        Returns True or False as to whether the test CNT intersects with any existing CNTs.
    '''
    intersect = False
    # if np.any(test_CNT.z < test_CNT.radius):
    #     intersect = True
    #     return intersect
    radius1 = test_CNT.radius
    for k in range(len(all_CNTs)):
        min_dist = all_CNTs[k].radius + radius1
        if _intersect(test_CNT.x, test_CNT.y, test_CNT.z, 
                      all_CNTs[k].x, all_CNTs[k].y, all_CNTs[k].z, 
                      min_dist, boxsize, boxsize_z):
            intersect = True
            return intersect
    return intersect


@njit()
def create_ball_coords(radius):
    """Generates a list of coordinates that are contained within a 3D sphere with a given radius.
    Parameters
    ----------
    radius : int32
        The radius of the sphere-shaped footprint.

    Returns
    -------
    ball_coords : List
        The footprint where elements of the neighborhood are 1 and 0 otherwise.
    """
    L = np.arange(-radius, radius + 1)
    ball_coords = List()
    for k in range(L.size):
        for j in range(L.size):
            for i in range(L.size):
                if (L[k]**2 + L[j]**2 + L[i]**2 < radius**2):
                    ball_coords.append((L[k], L[j], L[i]))
                
    return ball_coords


@njit()
def select_dilate_nb(xlist, ylist, zlist, coords, output_array, sizeXY, sizeZ, val):
    '''A dilation operation that only operates on the coordinates provided, instead of the whole array.
    Parameters
    ----------
    xlist : List[float64]
        List of x-positions.
    ylist : List[float64]
        List of y-positions.
    zlist : List[float64]
        List of z-positions.
    coords : List[int32]
        List of xyz coordinates over which to perform the dilation.
    output_array : ndarray
        Array upon which the dilation operation is performed.
    sizeXY : int32
        Size of box in x and y dimensions.
    sizeZ : int32
        Size of box in z dimension.
    val : int32
        Value to assign to the dilated coordinates.
    Returns
    -------
    None
    
    '''
    for i in range(xlist.shape[0]):
        for val_trio in coords:
            zcoord = int((val_trio[0]+round(zlist[i]))%sizeZ)
            ycoord = int((val_trio[1]+round(ylist[i]))%sizeXY)
            xcoord = int((val_trio[2]+round(xlist[i]))%sizeXY)
            output_array[zcoord, ycoord, xcoord] = val


def create_random_CNT(BoxXY, BoxZ, theta, psi, length, radius):
    '''Creates a single random CNT
    Parameters
    ----------
    BoxXY : int32
        Size of the box in the x and y dimensions.
    BoxZ : int32
        Size of the box in the z dimension.
    theta : float32
        Euler angle describing the MWCNT rotation about the Z-axis.
    psi : float32
        Euler angle describing the MWCNT rotation from vertical into the XY-plane (about the Y-axis).
    length : float32
        Length of the MWCNT.
    radius : float32
        Radius of the MWCNT.

    Returns
    -------
    random_CNT : CNT
        An instance of the CNT class created with random placement in the XYZ dimensions provided.
    '''
    r = np.linspace(-length/2, length/2, int(abs(length/2/radius+1)))
    x_center = np.random.random()*BoxXY
    y_center = np.random.random()*BoxXY
    z_center = np.random.random()*BoxZ
    x_test = (x_center + r*np.sin(theta)*np.cos(psi))%BoxXY
    y_test = (y_center + r*np.sin(theta)*np.sin(psi))%BoxXY
    z_test = (z_center + r*np.cos(theta))%BoxZ
    return CNT(x_center, y_center, z_center, 
                 x_test, y_test, z_test, psi, 
                 theta, length, radius)


def create_all_CNT(num_trials, radius_mu, radius_sigma, theta_mu, theta_sigma, length_lower, length_upper, BoxXY, BoxZ):
    '''Creates a full box of CNTs
    Parameters
    ----------
    num_trials : int32
        Number of test trials.
    radius_mu : float64
        Parameter mu for the log-normal distribution in CNT radius.
    radius_sigma : float64
        Parameter sigma for the log-normal distribution in CNT radius.
    theta_mu : float64
        Mean of the normal distribution for Euler angle theta.
    theta_sigma : float64
        Standard deviation of the normal distribution for Euler angle theta.
    length_lower : float64
        Lower limit of the uniform distribution for CNT length.
    length_upper : float64
        Upper limit of the uniform distribution for CNT length.
    BoxXY : int32
        Size of box in x and y dimensions.
    BoxZ : int32
        Size of box in z dimension.

    Returns
    -------
    all_CNTs : List[CNT]
        Returns list of all CNTs placed into the box.
    '''
    all_CNTs = List()
    psi = np.random.random()*np.pi*2
    theta = np.random.normal(theta_mu,theta_sigma)
    radius = np.random.lognormal(radius_mu, sigma=radius_sigma)
    # length= np.random.normal(length_mu, length_sigma)
    length = np.random.uniform(length_lower, length_upper)
    all_CNTs.append(create_random_CNT(BoxXY, BoxZ, theta, psi, length, radius))
    for i in range(num_trials):
        psi = np.random.random()*np.pi*2
        theta = np.random.normal(theta_mu,theta_sigma)
        radius = np.random.lognormal(radius_mu, sigma=radius_sigma)
        length = np.random.uniform(length_lower, length_upper)
        test_CNT = create_random_CNT(BoxXY, BoxZ, theta, psi, length, radius)
        if not test_intersect(test_CNT, all_CNTs, BoxXY, BoxZ):
            all_CNTs.append(test_CNT)
    return all_CNTs


def upscale_CNT(CNT,BoxXY, BoxZ):
    '''Upscales CNT central axis sampling for dilation operation. Replaces coordinates in place.
    Parameters
    ----------
    CNT : CNT
        Instance of CNT class.
    BoxXY : int32
        Size of box in x and y dimensions.
    BoxZ : int32
        Size of box in z dimension.

    Returns
    -------
    None
    '''
    r = np.linspace(-CNT.length/2, CNT.length/2, int(CNTs[i].length+1))
    CNT.x = (CNT.x_center + r*np.sin(CNT.theta)*np.cos(CNT.psi))%BoxXY
    CNT.y = (CNT.y_center + r*np.sin(CNT.theta)*np.sin(CNT.psi))%BoxXY
    CNT.z = (CNT.z_center + r*np.cos(CNT.theta))%BoxZ

@njit(parallel=True)
def dilate_all_par(CNTs, all_CNTs_dilated, BoxXY, BoxZ, val):
    '''Performs the dilation operation once to create a solid CNT. The first dilation operation assigns an arbitrary value to all voxels within the CNT radius. 
    Parameters
    ----------
    CNTs : List[CNT]
        List of CNTs
    all_CNTs_dilated : ndarray
        Array of zeros into which the dilation operation will be written.
    BoxXY : int32
        Size of box in x and y dimensions.
    BoxZ : int32
        Size of box in z dimension.
    fraction_hollow : float32
        Value between 0 and 1 denoting what fraction of the CNT radius is hollow.
    val : float32
        Value to assign to voxels.

    Returns
    -------
    None
    '''
    for i in prange(len(CNTs)):
        coords  = create_ball_coords(np.round(CNTs[i].radius))
        select_dilate_nb(CNTs[i].x, CNTs[i].y, CNTs[i].z, coords, all_CNTs_dilated, BoxXY, BoxZ, val)

@njit(parallel=True)
def dilate_all_thetapsi(CNTs, theta_dilated, psi_dilated, BoxXY, BoxZ):
    '''Dilation operation that assigns theta and psi to all voxels within the CNT radius. Used to demonstrate effects of changing uniaxial dielectric function alignment from radial to coaxial.
    Parameters
    ----------
    CNTs : List[CNT]
        List of CNTs
    theta_dilated : ndarray
        Array of zeros into which the dilation operation will be written for theta values.
    psi_dilated : ndarray
        Array of zeros into which the dilation operation will be written for psi values.
    BoxXY : int32
        Size of box in x and y dimensions.
    BoxZ : int32
        Size of box in z dimension.
    fraction_hollow : float32
        Value between 0 and 1 denoting what fraction of the CNT radius is hollow.

    Returns
    -------
    None
    '''
    for i in prange(len(CNTs)):
        coords  = create_ball_coords(np.round(CNTs[i].radius))
        select_dilate_nb(CNTs[i].x, CNTs[i].y, CNTs[i].z, coords, theta_dilated, BoxXY, BoxZ, CNTs[i].theta)
        select_dilate_nb(CNTs[i].x, CNTs[i].y, CNTs[i].z, coords, psi_dilated, BoxXY, BoxZ, CNTs[i].psi)
        

@njit(parallel=True)
def dilate_all_par_hollow(CNTs, all_CNTs_dilated, BoxXY, BoxZ, fraction_hollow, inside_value=0):
    '''Performs the dilation operation twice to create a hollow CNT. The first dilation operation assigns a value of 1 to all voxels within the CNT radius. 
    The second dilation operation assigns an arbitrary value to all voxels within some fraction of the CNT radius.
    Parameters
    ----------
    CNTs : List[CNT]
        List of CNTs
    all_CNTs_dilated : ndarray
        Array of zeros into which the dilation operation will be written.
    BoxXY : int32
        Size of box in x and y dimensions.
    BoxZ : int32
        Size of box in z dimension.
    fraction_hollow : float32
        Value between 0 and 1 denoting what fraction of the CNT radius is hollow.
    inside_value : float32
        Value to assign to the hollow interior of the CNT.

    Returns
    -------
    None
    '''
    for i in prange(len(CNTs)):
        radius_val = np.round(CNTs[i].radius)
        coords  = create_ball_coords(radius_val)
        select_dilate_nb(CNTs[i].x[radius_val:-radius_val], CNTs[i].y[radius_val:-radius_val], CNTs[i].z[radius_val:-radius_val], coords, all_CNTs_dilated, BoxXY, BoxZ, 1)
        coords = create_ball_coords(np.round(CNTs[i].radius*fraction_hollow))
        select_dilate_nb(CNTs[i].x, CNTs[i].y, CNTs[i].z, coords, all_CNTs_dilated, BoxXY, BoxZ, inside_value)


@njit(parallel=True)
def upscale_all_CNTs(CNTs, BoxXY, BoxZ):
    '''Upscales central axis of all CNTs for subsequent dilation operation. Replaces coordinates in place.
    Parameters
    ----------
    CNTs : List[CNT]
        List of CNTs
    BoxXY : int32
        Size of box in x and y dimensions.
    BoxZ : int32
        Size of box in z dimension.

    Returns
    -------
    None
    '''
    for i in prange(len(CNTs)):
        r = np.linspace(-CNTs[i].length/2, CNTs[i].length/2, int(CNTs[i].length+1))
        CNTs[i].x = (CNTs[i].x_center + r*np.sin(CNTs[i].theta)*np.cos(CNTs[i].psi))%BoxXY
        CNTs[i].y = (CNTs[i].y_center + r*np.sin(CNTs[i].theta)*np.sin(CNTs[i].psi))%BoxXY
        CNTs[i].z = (CNTs[i].z_center + r*np.cos(CNTs[i].theta))%BoxZ



def find_angles(CNTs, sigma=3):
    '''Determines the radial orientation of the uniaxial dielectric function at all points in the dilated array. Uses a Gaussian blur and gradient operation to find direction normal to all points on the CNT.
    Parameters
    ----------
    CNTs : List[CNT]
        List of CNTs.
    sigma : float32
        Standard deviation of the Gaussian used in the blurring operation.
    Returns
    -------
    theta : ndarray
        Array of theta values.
    psi : ndarray
        Array of psi values.
    '''
    try:
        blurred = gaussian_filter(CNTs, sigma=sigma)
    except TypeError:
        blurred = gaussian_filter(cp.asarray(CNTs), sigma=sigma)
    grad_z, grad_y, grad_x = cp.gradient(blurred)
    orient_mag = cp.sqrt(grad_x**2 + grad_y**2 + grad_z**2)
    grad_x /= -orient_mag 
    grad_y /= -orient_mag
    grad_z /= -orient_mag
    grad_x[cp.isinf(grad_x) | cp.isnan(grad_x)] = 0
    grad_y[cp.isinf(grad_y) | cp.isnan(grad_y)] = 0
    grad_z[cp.isinf(grad_z) | cp.isnan(grad_z)] = 0
    grad_z[grad_z > 1] = 1
    grad_z[grad_z < -1] = -1
    psi = cp.arctan2(grad_y, grad_x)
    theta = cp.arccos(grad_z)
    return theta, psi



import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib import rc, gridspec
import matplotlib
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import numpy as np
import h5py
import datetime
import warnings


def check_NumMat(f, morphology_type):
    """
    Checks and validates the number of materials in the HDF5 file

    Parameters
    ----------

    f : file object
    morphology_type : int
        0 - Euler, 1 - Vector

    Returns
    -------

    num_mat : int
        Number of materials in file

    """
    morphology_num = f['Morphology_Parameters/NumMaterial'][()]

    if morphology_type == 0:
        num_mat = 0
        while f'Euler_Angles/Mat_{num_mat + 1}_Vfrac' in f.keys():
            num_mat += 1
    elif morphology_type == 1:
        num_mat = 0
        while f'Vector_Morphology/Mat_{num_mat + 1}_unaligned' in f.keys():
            num_mat += 1

    assert morphology_num == num_mat, 'Number of materials does not match manual count of materials. Recheck hdf5'

    return num_mat


def readH5_vector(filename):
    """
    Reads in Vector Morphology from HDF5 file and converts to Euler Morphology. Checks that total material volume fractions sum to 1 for all voxels.

    Parameters
    ----------

    filename : str or path

    Returns
    -------
        Vfrac : ndarray
            Volume fraction of all materials. [Material, Z, Y, X]
        S : ndarray
            Fraction of alignment for each material. [Material, Z, Y, X]
        theta : ndarray
            Second Euler angle (ZYZ convention). [Material, Z, Y, X]
        psi : ndarray
            Third Euler angle (ZYZ convention). [Material, Z, Y, X]

    """
    with h5py.File(filename, 'r') as f:
        num_mat = check_NumMat(f, morphology_type=1)

        ds = f['Vector_Morphology/Mat_1_unaligned'][()]

        unaligned = np.zeros((num_mat, *ds.shape))
        alignment = np.zeros((num_mat, *ds.shape, 3))

        for i in range(0, num_mat):
            unaligned[i, ...] = f[f'Vector_Morphology/Mat_{i+1}_unaligned'][()]
            alignment[i, ...] = f[f'Vector_Morphology/Mat_{i+1}_alignment'][()]

        # calculate total material
        total_material = np.sum(unaligned, axis=0) + np.sum(alignment**2, axis=(0, -1))

    # assert that the entire morphology has total material equal to 1
    assert np.allclose(total_material, 1), 'Not all voxels in morphology have Total Material equal to 1'

    # convert vector to Euler for visualization purposes
    S = np.zeros((num_mat, *ds.shape))
    Vfrac = S.copy()
    theta = S.copy()
    psi = S.copy()

    s2 = np.sum(alignment**2, axis=-1)
    Vfrac = unaligned + s2
    S = s2/Vfrac
    np.nan_to_num(S, copy=False)
    S = np.round(S, 6)
    # calculate theta and psi from vectors
    with np.errstate(invalid='ignore'):
        normed_vectors = alignment/np.sqrt(s2[..., np.newaxis])
    np.nan_to_num(normed_vectors, copy=False)
    theta = np.arccos(normed_vectors[..., 2])
    psi = np.arctan2(normed_vectors[..., 1], normed_vectors[..., 0])

    return Vfrac, S, theta, psi


def readH5_euler(filename):
    """
    Reads in Euler Morphology from HDF5 file. Checks that total material volume fractions sum to 1 for all voxels.

    Parameters
    ----------

    filename : str or path

    Returns
    -------
        Vfrac : ndarray
            Volume fraction of all materials. [Material, Z, Y, X]
        S : ndarray
            Fraction of alignment for each material. [Material, Z, Y, X]
        theta : ndarray
            Second Euler angle (ZYZ convention). [Material, Z, Y, X]
        psi : ndarray
            Third Euler angle (ZYZ convention). [Material, Z, Y, X]

    """
    with h5py.File(filename, 'r') as f:
        num_mat = check_NumMat(f, morphology_type=0)

        ds = f['Euler_Angles/Mat_1_Vfrac'][()]

        Vfrac = np.zeros((num_mat, *ds.shape))
        S = Vfrac.copy()
        theta = Vfrac.copy()
        psi = Vfrac.copy()

        # 'Mat_1_Psi', 'Mat_1_S', 'Mat_1_Theta', 'Mat_1_Vfrac'

        for i in range(0, num_mat):
            Vfrac[i, :, :, :] = f[f'Euler_Angles/Mat_{i+1}_Vfrac']
            S[i, :, :, :] = f[f'Euler_Angles/Mat_{i+1}_S']
            theta[i, :, :, :] = f[f'Euler_Angles/Mat_{i+1}_Theta']
            psi[i, :, :, :] = f[f'Euler_Angles/Mat_{i+1}_Psi']
            psi = np.mod(psi, np.pi)
            # calculate total material
            total_material = np.sum(Vfrac, axis=0)

    # assert that the entire morphology has total material equal to 1
    assert np.allclose(total_material, 1), 'Not all voxels in morphology have Total Material equal to 1' 

    return Vfrac, S, theta, psi


def checkH5(filename='perp82.hd5', z_slice=0, subsample=None, outputmat=None, runquiet=False, plotstyle='light'):
    """
    Reads in morphology HDF5 file and checks that the format is consistent for CyRSoXS. Optionally plots and returns select quantities.

    Parameters
    ----------

        filename : str or path
            Name of HDF5 morphology file to check
        z_slice : int
            Which z-slice of the array to plot.
        subsample : int
            Number of voxels to display in X and Y
        outputmat : int
            Number of which material to return
        runquiet : bool
            Boolean flag for running without plotting or outputting to console
        plotstyle : str
            Use a light or dark background for plots. 'dark' - dark, 'light' - light

    Returns
    -------
        Vfrac : ndarray
            Volume fraction of selected material. [ Z, Y, X]
        S : ndarray
            Fraction of alignment for selected material. [Z, Y, X]
        theta : ndarray
            Second Euler angle (ZYZ convention). [Z, Y, X]
        psi : ndarray
            Third Euler angle (ZYZ convention). [Z, Y, X]
    """
    style_dict = {'dark': 'dark_background',
                  'light': 'default'}

    begin_time = datetime.datetime.now()

    with h5py.File(filename, 'r') as f:
        # check morphology type
        if 'Euler_Angles' in f.keys():
            morphology_type = 0
        elif 'Vector_Morphology' in f.keys():
            morphology_type = 1
        else:
            raise KeyError('Neither \"Euler_Angles\" or \"Vector_Morphology\" group detected in hdf5')
            # morphology_type = None

    if morphology_type == 0:
        Vfrac, S, theta, psi = readH5_euler(filename)
        num_mat, zdim, ydim, xdim = Vfrac.shape
    elif morphology_type == 1:
        Vfrac, S, theta, psi = readH5_vector(filename)
        num_mat, zdim, ydim, xdim = Vfrac.shape

    # check to make sure arrays are float
    improper_types = []
    for array, array_name in zip([Vfrac, S, theta, psi], ['Vfrac', 'S', 'theta', 'psi']):
        if 'float' not in array.dtype.name:
            warnings.warn(f'{array_name} dtype is {array.dtype.name} and incompatible with CyRSoXS. Array dtype must be some variant of float')
            improper_types.append([array_name, array.dtype.name])

    # raise exception after going through all material arrays so user can fix all of them at once
    if len(improper_types) != 0:
        for val in improper_types:
            print(val[0], val[1])  
        raise TypeError('One or more of the material arrays dtype is not float. See above for details')

    if subsample is None:
        subsample = ydim # y dimension

    if not runquiet:
        print(f'Dataset dimensions (Z, Y, X): {zdim} x {ydim} x {xdim}')
        print(f'Number of Materials: {num_mat}')
        print('')

    if z_slice > (zdim-1):
        print(f'Error: z_slice of {z_slice} is greater than the maximum index of {zdim-1}. Using z_slice = {zdim-1} instead.')
        z_slice = zdim-1

    if not runquiet:

#         if plotstyle == 'dark':
        plt.style.use(style_dict[plotstyle])
        font = {'family': 'sans-serif',
                'sans-serif': 'DejaVu Sans',
                'weight': 'regular',
                'size': 18}

        rc('font', **font)

        for i in range(0,num_mat):    
            print(f'Material {i+1} Vfrac. Min: {np.amin(Vfrac[i,:,:,:])} Max: {np.amax(Vfrac[i,:,:,:])}')
            print(f'Material {i+1} S. Min: {np.amin(S[i,:,:,:])} Max: {np.amax(S[i,:,:,:])}')
            print(f'Material {i+1} theta. Min: {np.amin(theta[i,:,:,:])} Max: {np.amax(theta[i,:,:,:])}')
            print(f'Material {i+1} psi. Min: {np.amin(psi[i,:,:,:])} Max: {np.amax(psi[i,:,:,:])}')

            fig = plt.figure(figsize=(22, 33))

            gs = gridspec.GridSpec(nrows=5,
                                   ncols=2,
                                   figure=fig,
                                   width_ratios=[1, 1],
                                   height_ratios=[3, 1, 0.1, 3, 1],
                                   wspace=0.3,
                                   hspace=0.2)

            start = int(ydim/2)-int(subsample/2)
            end = int(ydim/2)+int(subsample/2)

            ax1 = plt.subplot(gs[0, 0])
            Vfracplot = ax1.imshow(Vfrac[i, z_slice, start:end, start:end], cmap=plt.get_cmap('winter'), origin='lower', interpolation='none')
            ax1.set_ylabel('Y index')
            ax1.set_xlabel('X index')

            ax1.set_title(f'Material {i+1} Vfrac')
            Vfrac_cbar = plt.colorbar(Vfracplot, fraction=0.040)
            Vfrac_cbar.set_label('Vfrac: volume fraction', rotation=270, labelpad=22)

            ax2 = plt.subplot(gs[0, 1])
            Splot = ax2.imshow(S[i, z_slice, start:end, start:end], cmap=plt.get_cmap('nipy_spectral'), origin='lower', interpolation='none')
            ax2.set_ylabel('Y index')
            ax2.set_xlabel('X index')
            ax2.set_title(f'Material {i+1} S')
            S_cbar = plt.colorbar(Splot, fraction=0.040)
            S_cbar.set_label('S: orientational order parameter', rotation=270, labelpad=22)

            ax3 = plt.subplot(gs[1, 0])
            ax3.hist(Vfrac[i, :, :, :].flatten())
            ax3.set_title(f'Material {i+1} Vfrac')
            ax3.set_xlim(left=0)
            ax3.set_xlabel('Vfrac: volume fraction')
            ax3.set_ylabel('num voxels')
            ax3.set_yscale('log')

            ax4 = plt.subplot(gs[1, 1])
            ax4.hist(S[i, :, :, :].flatten())
            ax4.set_title(f'Material {i+1} S')
            ax4.set_xlim(left=0)
            ax4.set_xlabel('S: orientational order parameter')
            ax4.set_ylabel('num voxels')
            ax4.set_yscale('log')

            ax5 = plt.subplot(gs[3, 0])
            norm = matplotlib.colors.Normalize(vmin=0, vmax=np.pi, clip=False)
            thetaplot = ax5.imshow(np.ma.masked_array(theta[i, z_slice, start:end, start:end], np.logical_or(Vfrac[i, z_slice, start:end, start:end] < 0.01, S[i, z_slice, start:end, start:end] < 0.01)),
                                   cmap=plt.get_cmap('jet'), norm=norm, origin='lower', interpolation='none')
            ax5.set_ylabel('Y index')
            ax5.set_xlabel('X index')            
            ax5.set_title(f'Material {i+1} theta')
            theta_cbar = plt.colorbar(thetaplot, fraction=0.040)
            theta_cbar.set_label('theta in radians', rotation=270, labelpad=22)

            ax6 = plt.subplot(gs[3, 1])
            norm = matplotlib.colors.Normalize(vmin=0, vmax=np.pi, clip=False)
            psiplot = ax6.imshow(np.ma.masked_array(psi[i, z_slice, start:end, start:end], np.logical_or(Vfrac[i, z_slice, start:end, start:end] < 0.01, S[i, z_slice, start:end, start:end] < 0.01)),
                                 cmap=plt.get_cmap('hsv'), norm=norm, origin='lower', interpolation='none')
            ax6.set_ylabel('Y index')
            ax6.set_xlabel('X index')            
            ax6.set_title(f'Material {i+1} psi')
            psi_cbar = plt.colorbar(psiplot, fraction=0.040)
            psi_cbar.set_label('psi in radians', labelpad=22)       

            ax7 = plt.subplot(gs[4, 0])
            ax7.hist(theta[i, :, :, :].flatten())
            ax7.set_title(f'Material {i+1} theta')
            ax7.set_xlim(left=0)
            ax7.set_xlabel('theta in radians')
            ax7.set_ylabel('num voxels')
            ax7.set_yscale('log')

            ax8 = plt.subplot(gs[4, 1])
            ax8.hist(psi[i, :, :, :].flatten())
            ax8.set_title(f'Material {i+1} psi')
            ax8.set_xlim(left=0)
            ax8.set_xlabel('psi in radians')
            ax8.set_ylabel('num voxels')
            ax8.set_yscale('log')

            ax4i = inset_axes(ax6,  axes_class=matplotlib.projections.get_projection_class('polar'), width=2, height=2, axes_kwargs={"alpha": 0})

            ax4i.grid(False)
            # creates inset legend for orientation
            azimuths = np.deg2rad(np.arange(0, 360, 1))
            zeniths = np.linspace(5, 10, 50)
            values = np.mod(azimuths, np.pi) * np.ones((50, 360))
            ax4i.plot(0, 0, 'o', ms=75, mec='none', mfc=plt.rcParams['axes.facecolor'], mew=2, zorder=0, alpha=0.7)
            ax4i.pcolormesh(azimuths, zeniths, values, cmap=cm.hsv, shading='auto')
            ax4i.set_axis_off()
            ax4i.arrow(0, 0, 0, 4, width=0.015, head_width=0.25, head_length=0.5)  #,facecolor='k',edgecolor='k')
            ax4i.arrow(np.pi/2, 0, 0, 4, width=0.015, head_width=0.25, head_length=0.5)  #,facecolor='k',edgecolor='k')
            ax4i.text(2*np.pi-np.deg2rad(40), 3.5, 'X')
            ax4i.text(np.pi/2+np.deg2rad(40), 3.5, 'Y')

            ax5i = inset_axes(ax5,  axes_class=matplotlib.projections.get_projection_class('polar'), width=2, height=2, axes_kwargs={"alpha": 0})

            ax5i.grid(False)
            # creates inset legend for orientation
            azimuths_t = np.deg2rad(np.arange(-90, 90, 1))
            zeniths_t = np.linspace(5, 10, 50)
            values_t = np.mod(np.pi/2-azimuths_t, np.pi) * np.ones((50, 180))
            ax5i.plot(0, 0, 'o', ms=75, mec='none', mfc=plt.rcParams['axes.facecolor'], mew=2, zorder=0, alpha=0.7)
            ax5i.pcolormesh(azimuths_t, zeniths_t, values_t, cmap=cm.jet, shading='auto')
            ax5i.set_axis_off()
            ax5i.arrow(0, 0, 0, 4, width=0.015, head_width=0.25, head_length=0.5)  #,facecolor='k',edgecolor='k')
            ax5i.arrow(np.pi/2, 0, 0, 4, width=0.015, head_width=0.25, head_length=0.5)  #, facecolor='k',edgecolor='k')
            ax5i.text(2*np.pi-np.deg2rad(40), 3.5, 'X')
            ax5i.text(np.pi/2+np.deg2rad(40), 3.5, 'Z')

            plt.show()

    plt.rcParams.update(plt.rcParamsDefault)

    if not runquiet:
        print(f'Total Vfrac whole model. Min: {np.amin(np.sum(Vfrac, axis=0))} Max: {np.amax(np.sum(Vfrac, axis=0))}')
        print(datetime.datetime.now() - begin_time)
    if not (outputmat is None):
        start = int(ydim/2)-int(subsample/2)
        end = int(ydim/2)+int(subsample/2)
        return Vfrac[outputmat, z_slice, start:end, start:end], S[outputmat, z_slice, start:end, start:end], psi[outputmat, z_slice, start:end, start:end], theta[outputmat, z_slice, start:end, start:end]



import math
import numpy as np
import os
import h5py
from lib.util import *
from lib.generateConstants import *
from lib.reduce import *
import sys
import subprocess
import pathlib
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import xarray as xr
from scipy.optimize import curve_fit
plt.rcParams.update({'font.size':10})

class Config:

    def __init__(self, size_xy: int = None, size_z: int = None, phys_size: float = None,
                a_nm: float = None, b_nm: float = None, theta: float = None, radius_nm: float = None):

        # Simulation Parameters
        self.size_xy = size_xy
        self.size_z  = size_z
        self.phys_size   = phys_size

        # Materials Parameters
        self.num_materials = 2

        # Lattice Parameters
        self.a_nm  = a_nm   # Unit cell length a in nm
        self.b_nm  = b_nm   # Unit cell length b in nm
        self.theta = theta  # Angle between the unit cell edges a and b
        self.radius_nm = radius_nm    # Radius of the cylinders in nm

class Morphology:
    def __init__(self, config: Config = None):
        self.cfg = config
        self.box = None
        self.coords = None
        self.circle_tile = None

    def makeMorphology(self):
        # Initialize box
        self.box = generateBox(self.cfg)
        # Generate lattice coordinates
        self.coords = generateCoords(self.cfg)
        # Populate the box
        self.box = populateBox(self)
        # Generate the material maps that will be used for the Cy-RSoXS input
        self.unaligned, self.aligned = generateMaterialMaps(self.box, self.cfg.num_materials)

def generateBox(cfg):
    '''Generate the simulation box.'''
    box = np.zeros((cfg.size_xy, cfg.size_xy, cfg.size_z, cfg.num_materials + 1))
    box[:, :, :, 0] = 1
    return box

def generateCoords(cfg):

    def calcRange(size_xy, a_vector, b_vector, angle, range_overshoot):
        cosang = np.cos(np.deg2rad(angle))
        sinang = np.sin(np.deg2rad(angle))

        m_range = int(round(range_overshoot*(size_xy*cosang + size_xy*sinang)/np.amax(np.absolute(b_vector))))
        n_range = int(round(range_overshoot*(size_xy*sinang + size_xy*cosang)/np.amax(np.absolute(a_vector))))

        return m_range, n_range

    # Convert a and b lattice edge lengths into voxel lengths (integers) based on the phys_size
    a_voxels = int(round(cfg.a_nm/cfg.phys_size))
    b_voxels = int(round(cfg.b_nm/cfg.phys_size))

    # The lattice point coordinates are generated by looping over integer multiples of the lattice vectors (n*a + m*b)
    # The ranges of n and m are estimated by dividing the simulation box size by the a and b length scales
    # Lattices with non-90deg angles between a and b require overshooting the typical counting to fill the simulation box
    coords = []
    range_overshoot = 1.1

    b_0 = b_voxels*np.cos(np.deg2rad(cfg.theta))
    b_1 = b_voxels*np.sin(np.deg2rad(cfg.theta))

    a_vector = [a_voxels, 0]
    b_vector = [b_0, b_1]

    angle = 0
    origin = int(round(cfg.size_xy/2))

    m_range, n_range = calcRange(cfg.size_xy, a_vector, b_vector, angle, range_overshoot)

    # Coordinates are generated from -n to n and -m to m
    # Only coordinates inside the simulation box are added to the list
    for m in range(-m_range, m_range + 1):
        for n in range(-n_range, n_range + 1):

            new_coord_x = int(round(origin + n*a_vector[0] + m*b_vector[0]))
            new_coord_y = int(round(origin + n*a_vector[1] + m*b_vector[1]))

            if new_coord_x in range(0, cfg.size_xy) and new_coord_y in range(0, cfg.size_xy):
                coords.append([new_coord_x, new_coord_y, 1])

    return coords

def generateCircleTile(cfg):

    def getMajorityMaterialID(fractions):
        max_fraction = np.amax(fractions)
        material_ID  = np.where(fractions == max_fraction)
        return material_ID[0]

    radius = int(round(cfg.radius_nm/cfg.phys_size))
    width = int(round(2 * radius))
    circle_tile = np.zeros((width, width, cfg.num_materials + 1))
    circle_tile[:, :, 0] = 1
    origin_x, origin_y = np.size(circle_tile, 0)/2, np.size(circle_tile, 1)/2
    get_dist = lambda i, j: math.sqrt((i + 0.5 - origin_x)**2 + (j + 0.5 - origin_y)**2)

    for i in range(0, np.size(circle_tile, 0)):
        for j in range(0, np.size(circle_tile, 1)):

            r = get_dist(i, j)

            if r >= radius:
                f_B = 0
            else:
                f_B = 1

            f_A = 1 - f_B
            maj_material_ID = getMajorityMaterialID([f_A, f_B])
            circle_tile[i, j, :] = [f_A, f_B, maj_material_ID]

    return circle_tile

def populateBox(morphology):

    box = morphology.box
    coords = morphology.coords
    circle_tile = generateCircleTile(morphology.cfg)
    morphology.circle_tile = circle_tile

    def insertCircleTile(circle_tile, center, z):
        width = np.size(circle_tile, 0)
        minx = int(round(center[0] - width/2))
        miny = int(round(center[1] - width/2))
        maxx = int(round(minx + width))
        maxy = int(round(miny + width))

        # Check if cylinder is out of simulation bounds
        if minx >=0 and miny >=0 and maxx < len(box) and maxy < len(box):
            for i in range(maxx-minx):
                for j in range(maxy-miny):
                    if circle_tile[i, j, 0] != 1:
                        box[i+minx, j+miny, z, :] = circle_tile[i, j, :]

        return box

    for coord in coords:
        box = insertCircleTile(circle_tile, coord, 0)

    # We created the morphology as [X, Y, Z] but Cy-RSoXS takes [Z, Y, X]
    # Swap axes of our box
    box = np.swapaxes(box, 0, 2)
    return box

def generateMaterialMaps(box, num_materials):

    dimZ = np.size(box, 0)
    dimY = np.size(box, 1)
    dimX = np.size(box, 2)

    mat1_unaligned = box[:, :, :, 0]
    mat1_aligned = np.zeros((dimZ, dimY, dimX, 3))
    mat2_unaligned = box[:, :, :, 1]
    mat2_aligned = np.zeros((dimZ, dimY, dimX, 3))

    # The material maps define the fractional composition of material x in a given voxel
    # The element-wise sum of these arrays should give an array of all 1's
    # Currently, this check sums the unaligned maps because we do not support aligned fractions yet
    unaligned_sum = mat1_unaligned + mat2_unaligned
    unaligned_sum = unaligned_sum.flatten()

    if np.all(abs(unaligned_sum - 1) <= 1e-6) == False:
        raise ValueError('Uh oh! Not every voxel sums to 1')

    return [mat1_unaligned, mat2_unaligned], [mat1_aligned, mat2_aligned]

def generateHDF5(morphology, filepath: str = None, filename: str = None):

    f = h5py.File(filepath + filename + '.h5', 'w')

    g0   = f.create_group('Morphology_Parameters')
    dset_1 = g0.create_dataset('PhysSize', data = morphology.cfg.phys_size, dtype = np.float64)
    dset_2 = g0.create_dataset('NumMaterial', data = morphology.cfg.num_materials, dtype = np.float64)

    g1 = f.create_group('Vector_Morphology')
    dset_m1u = g1.create_dataset('Mat_1_unaligned', data=(morphology.unaligned[0]), dtype = np.float32, compression='gzip', compression_opts=9)
    g1['Mat_1_unaligned'].dims[0].label = 'Z'
    g1['Mat_1_unaligned'].dims[1].label = 'Y'
    g1['Mat_1_unaligned'].dims[2].label = 'X'
    dset_m2u = g1.create_dataset('Mat_2_unaligned', data=(morphology.unaligned[1]), dtype = np.float32, compression='gzip', compression_opts=9)
    g1['Mat_2_unaligned'].dims[0].label = 'Z'
    g1['Mat_2_unaligned'].dims[1].label = 'Y'
    g1['Mat_2_unaligned'].dims[2].label = 'X'

    dset_m1a = g1.create_dataset("Mat_1_alignment", data = (morphology.aligned[0]), dtype = np.float32, compression='gzip', compression_opts=9)
    g1['Mat_1_alignment'].dims[0].label = 'Z'
    g1['Mat_1_alignment'].dims[1].label = 'Y'
    g1['Mat_1_alignment'].dims[2].label = 'X'
    dset_m2a = g1.create_dataset("Mat_2_alignment", data = (morphology.aligned[1]), dtype = np.float32, compression='gzip', compression_opts=9)
    g1['Mat_2_alignment'].dims[0].label = 'Z'
    g1['Mat_2_alignment'].dims[1].label = 'Y'
    g1['Mat_2_alignment'].dims[2].label = 'X'

    f.close()

    
def generate_circle_lattice_test(return_folderpath=False, slurm_file=0, **slurm_kwargs):
    # default input parameters
    size_xy = 1024
    size_z = 1
    PhysSize = 2
    a_nm = 60
    b_nm = 60
    theta = 120
    radius_nm = 20
    NumMat = 2
    
    config = Config(size_xy = size_xy,
                size_z = size_z,
                phys_size = PhysSize,
                a_nm = a_nm,
                b_nm = a_nm,
                theta = theta,
                radius_nm = radius_nm)

    morphology = Morphology(config = config)
    morphology.makeMorphology()
    
    filepath = os.getcwd() + '/'
    filename = 'circle-lattice-test'
    generateHDF5(morphology, filepath=filepath, filename=filename)


    # write config and material files
    startEnergy = 285.0
    endEnergy = 286.0
    incrementEnergy = 1.0

    numThreads = 4;        #number of threads for execution
    #Files corresponding to Each material.
    dict={'Material0':'PEOlig2018.txt',
          'Material1':'vacuum'}


    # Label of energy to look for
    labelEnergy={"BetaPara":0,
                 "BetaPerp":1,
                 "DeltaPara":2,
                 "DeltaPerp":3,
                 "Energy":6}


    write_config2D(size_xy,PhysSize,[startEnergy,endEnergy],euler=False)

    # renamed "main" in generateConstants.py to something more descriptive
    write_materials(startEnergy,endEnergy,incrementEnergy,dict,labelEnergy,len(dict))

    # Create folder stamped with current datetime
    time_now = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    folder_name = 'circle_lattice_test_' + time_now
    folder_path = filepath + folder_name + '/'
    os.mkdir(folder_name)
    filename = filename + '.h5'
    subprocess.run(["mv","Material1.txt",f"./{folder_name}"])
    subprocess.run(["mv","Material2.txt",f"./{folder_name}"])
    subprocess.run(["mv","config.txt",f"./{folder_name}"])
    subprocess.run(["mv",f"{filename}",f"./{folder_name}"])

    if slurm_file == 1:
        write_slurm(filename, **slurm_kwargs)
        subprocess.run(["mv","job.slurm",f"./{folder_name}"])

    if return_folderpath:
        return folder_path
    
def run_slurm(folder_path):
    os.chdir(folder_path)
    subprocess.run(['sbatch','job.slurm'])
    

def test_circle_lattice(folder_path, savepng=False):
    h5_path = pathlib.Path(folder_path,'HDF5')
    h5s = sorted(list(h5_path.glob('*h5')))
    data = read_img_h5(h5s[0], PhysSize=2)
    remeshed_data = remesh(data)
    
    I_simulation = remeshed_data.mean('chi')
    
    # reference peak positions
    d_ref = 51.9615243
    q_star = 2*np.pi/d_ref
    n = np.array([1,3,4,7,9,12,13]) # first 7 peak orders
    q_ref = q_star*np.sqrt(n.astype(float))
    q_fit = []
    opt_params = []
    for peak in q_ref:
        q_lower = peak*0.9
        q_upper = peak*1.1
        I_fit = I_simulation.where((I_simulation.q > q_lower) & (I_simulation.q < q_upper) & ~np.isnan(I_simulation.values),drop=True)
#         print(I_fit.values)
        popt, pcov = curve_fit(gauss_bck, I_fit.q, I_fit.values,p0=(1e-2,peak,0.005), bounds=((0,q_lower,0),(10,q_upper,1)))
        q_fit.append(popt[1])
        opt_params.append(popt)
    
    q_fit = np.array(q_fit)
    
    assert(np.allclose(np.round(q_ref,3), np.round(q_fit,3)))
    
    

    if savepng:
        peaks_str = [r'q$^*$', r'$\sqrt{3}q^*$', r'$\sqrt{4}q^*$', 
                     r'$\sqrt{7}q^*$', r'$\sqrt{9}q^*$', 
                     r'$\sqrt{12}q^*$', r'$\sqrt{13}q^*$']
        centering = ['center','center','center','center','center','right','left']
        v_centering = ['center','center','center','center','center','top','bottom']
        with h5py.File(pathlib.Path(folder_path,'circle-lattice-test.h5'),'r') as f:
            model = f['Vector_Morphology']['Mat_2_unaligned'][()]
    
        plt.figure(constrained_layout=True,figsize=(3.25,2.7))
        plt.imshow(model[0,:,:],origin='lower',extent=[0,model.shape[1]*2,0,model.shape[2]*2])
        plt.xlabel('X [nm]')
        plt.ylabel('Y [nm]')
        plt.xlim(0,1000)
        plt.ylim(0,1000)
        plt.savefig(pathlib.Path(folder_path,'Model.png'),format='png',dpi=300,bbox_inches='tight')
        plt.close()
    

        plt.figure(constrained_layout=True,figsize=(3.25,2.7))
        I_simulation.plot(yscale='log')
        for i,peak in enumerate(q_ref):
            q_lower = peak*0.9
            q_upper = peak*1.1
            xnew = np.linspace(q_lower,q_upper)
            params = opt_params[i]
#             print(params,type(params))
            plt.plot(xnew,gauss_bck(xnew,*params)+1e-8,'k--')
            plt.text(peak,gauss_bck(peak,*params)*1.5,peaks_str[i],ha=centering[i],fontsize=6)
#             plt.plot(xnew,bck(xnew,*opt_params[i][-2:]),'k',linestyle='--')
        plt.xlabel(r'q [nm$^{-1}$]')
        plt.ylabel('I(q)')
        plt.xlim(4e-2,0.5)
        plt.ylim(1e-8)
        plt.savefig(pathlib.Path(folder_path,'PeakComparison.png'),format='png',dpi=300,bbox_inches='tight')
        plt.close()

        plt.figure(figsize=(3.25,2.7),constrained_layout=True)
        remeshed_data.plot(norm=LogNorm(1e-6,1e0))
        plt.title('Remeshed CyRSoXS Output')
        plt.xlabel(r'q [nm$^{-1}$]')
        plt.ylabel(r'$\chi$ [Degrees]')
        plt.savefig(pathlib.Path(folder_path,'remeshed.png'),format='png',dpi=300,bbox_inches='tight')
        plt.close()

        plt.figure(figsize=(3.25,2.7),constrained_layout=True)
        data.plot(norm=LogNorm(1e-6,1e0),cmap='jet')
#         plt.title('2D CyRSoXS Output')
        plt.xlabel(r'q$_x$ [nm$^{-1}$]')
        plt.ylabel(r'q$_y$ [nm$^{-1}$]')
        plt.savefig(pathlib.Path(folder_path,'raw.png'),format='png',dpi=300,bbox_inches='tight')
        plt.close()
        
        plt.figure(figsize=(3.25,2.7),constrained_layout=True)
        data.plot(norm=LogNorm(1e-6,1e0),cmap='jet')
        plt.xlim(0,0.5)
        plt.ylim(0,0.5)
#         plt.title('2D CyRSoXS Output')
        plt.xlabel(r'q$_x$ [nm$^{-1}$]')
        plt.ylabel(r'q$_y$ [nm$^{-1}$]')
        for i,peak in enumerate(q_ref):
            plt.text(0.03,peak,peaks_str[i],va=v_centering[i],ha='center',fontsize=6)
        plt.savefig(pathlib.Path(folder_path,'raw_zoomed.png'),format='png',dpi=300,bbox_inches='tight')
        plt.close()
        
    
   
    
    
def gauss_bck(x,a,b,c):
    return a*np.exp(-(x-b)**2/c**2)

def bck(x,d,e):
    return d*x + e
    
    
    
    
    
if __name__ == '__main__':
    version = '1.0.1.1'
    slurm_file = 0
    
    for arg in sys.argv:
        if arg.startswith('slurm'):
            slurm_file = int(arg.split('=')[-1])
        if arg.startswith('version'):
            version = arg.split('=')[-1]

    folderpath = generate_circle_lattice_test(return_folderpath=True,slurm_file=slurm_file,cyrsoxs_version=version)
    
    if slurm_file == 1:
        import time
        run_slurm(folderpath)
        hdf5path = pathlib.Path(folderpath,'HDF5')
        while len(list(hdf5path.glob('*h5'))) < 1:
            time.sleep(1)
        test_circle_lattice(folderpath,savepng=True)





from enum import Enum
import numpy as np
from NRSS.morphology import Material, Morphology
from PyHyperScattering.integrate import WPIntegrator
import pickle, lzma
from pathlib import Path
import glob
from tqdm import tqdm
from pprint import pprint
from matplotlib import colormaps, gridspec, rc
from matplotlib import pyplot as plt

# FROM NOTEBOOK 2

# class syntax
class EulerStyle(Enum):
    """
    a class for qualitative model changes to Euler angle style relative to particle
    """
    RADIAL = 1
    """
    creates extraordindary index orientation that is radial to the core particle center
    """
    TANGENTIAL_LAT = 2
    """
    creates extraordindary index orientation that is tangential to the core particle center and latitudinal such that all orientations are parallel to the XY plane
    """
    TANGENTIAL_LONG = 3
    """
    creates extraordindary index orientation that is tangential to the core particle center and longitudinal such that all orientations are not parallel to the XY plane except at the north and south poles
    """
def adsorbed_polymer_morphology(args):
    """
    Generates a morphology object for an adsorbed polymer on a sphere.

    Args:
        args (dict): A dictionary containing the following keys:
            - "vd" (int): The vertical dimension of the grid.
            - "ld" (int): The lateral dimension of the grid.
            - "radius_nm" (float): The radius of the sphere in nanometers.
            - "PhysSize_nm_per_voxel" (float): The physical size of each voxel in nanometers.
            - "euler_style" (str): The style of Euler angles to use. From an enumerated type: EulerStyle.RADIAL, EulerStyle.TANGENTIAL_LAT, or EulerStyle.TANGENTIAL_LONG.
            - "S0" (float): The initial value of the S-field next to the nanoparticle.
            - "S_slope_per_nm" (float): The slope of the S-field in nanometers.
            - "energies" (list): A list of energies for the model.
            - "oc_polymer" (object): The opt_constants object for the polymer.
            - "oc_particle" (object): The opt_constants object for the particle.

    Returns:
        morph (Morphology): A Morphology object representing the adsorbed polymer on a sphere.
    """
    # make grid
    z, y, x = np.ogrid[0 : args["vd"], 0 : args["ld"], 0 : args["ld"]]
    # distance from center
    dist_from_ctr = np.sqrt(
        (x - x.mean()) ** 2
        + (y - y.mean()) ** 2
        + (z - z.mean()) ** 2
    )
    # boolean sphere
    v_sphere = dist_from_ctr < args["radius_nm"] / args["PhysSize_nm_per_voxel"]

    # Euler angle math
    delta_x = x - args["ld"] / 2
    delta_y = y - args["ld"] / 2
    delta_z = z - args["vd"] / 2

    # the psi and theta for the radial case
    psi = np.arctan2(delta_y, delta_x) + 0 * delta_z
    theta = np.arctan2(np.sqrt(delta_x**2 + delta_y**2), delta_z)

    # match - case is similar to a series of if - elseif statements
    match args["euler_style"]:
        case EulerStyle.RADIAL:
            pass
            # do nothing because psi and theta are already correct for radial
        case EulerStyle.TANGENTIAL_LAT:
            # modify the psi and theta for tangential_lat
            theta = np.full_like(theta, np.pi / 2)
            psi = psi - np.pi / 2
        case EulerStyle.TANGENTIAL_LONG:
            # modify the psi and theta for tangential_lon
            psi = psi
            theta = np.where(theta < np.pi / 2, theta + np.pi / 2, theta - np.pi / 2)
        case _:
            # raise a value error if this doesn't come in as one of the enumerated types
            raise ValueError("invalid euler style")
        
    # S-field math
    S_field = (
        args["S0"]
        + (dist_from_ctr - args["radius_nm"] / args["PhysSize_nm_per_voxel"])
        * args["S_slope_per_nm"]
    )
    S_field = np.maximum(0, S_field)
    S_field = np.minimum(args["S0"], S_field)
    S_field *= v_sphere == False

    # constructing the model object
    a_zeros = np.zeros(v_sphere.shape, dtype=np.float32)

    mat1_poly = Material(
        materialID=1,
        Vfrac=(1 - v_sphere).astype(np.float32),
        S=S_field.astype(np.float32),
        theta=theta.astype(np.float32),
        psi=psi.astype(np.float32),
        NumZYX=v_sphere.shape,
        energies=args["energies"],
        opt_constants=args["oc_polymer"].opt_constants,
        name="polymer",
    )

    mat2_particle = Material(
        materialID=2,
        Vfrac=v_sphere.astype(np.float32),
        S=a_zeros,
        theta=a_zeros,
        psi=a_zeros,
        NumZYX=v_sphere.shape,
        energies=args["energies"],
        opt_constants=args["oc_particle"].opt_constants,
        name="particle",
    )

    morph = Morphology(
        2,
        {1: mat1_poly, 2: mat2_particle},
        PhysSize=args["PhysSize_nm_per_voxel"],
    )
    # the return statement is what the function delivers as an output
    return morph

# FROM NOTEBOOK 2
def adsorbed_polymer_run(args):
    """
    Run the adsorbed polymer model.

    Args:
        args (dict): A dictionary containing the arguments for running the model.

    Returns:
        results (dict): A dictionary containing the results of the model run. The dictionary has the following keys:
            - "I" (numpy.ndarray): The mean of the remeshed data along the "chi" dimension.
            - "A" (numpy.ndarray): The result of the rsoxs.AR function applied to the remeshed data with a chi_width of 45.
            - "visualizations" (dict): A dictionary containing the visualizations created by the morphology.visualize_materials function.
            - "args" (dict): A copy of the input arguments.

    Description:
        This function runs the adsorbed polymer model using the provided arguments. It creates a morphology object using the adsorbed_polymer_morphology function and then runs the model using the morphology.run method. The resulting data is then remeshed using the WPIntegrator.integrateImageStack method. The mean of the remeshed data along the "chi" dimension is calculated and stored in the "I" key of the results dictionary. The result of the rsoxs.AR function applied to the remeshed data with a chi_width of 45 is stored in the "A" key of the results dictionary. Visualizations of the model are created using the morphology.visualize_materials function and stored in the "visualizations" key of the results dictionary. The input arguments are stored in the "args" key of the results dictionary. The results dictionary is then returned.
    """    
    # notice how we're passing args to the adsorbed_polymer_morphology function
    # we could also have args that affect how the model is run (but we don't right now)

    # we start right away with the adsorbed_polymer_morphology function
    # Create morphology object
    morph = adsorbed_polymer_morphology(args)

    # FFT "window" parameter
    morph.inputData.windowingType = 0  # cy.FFTWindowing.Hanning


    # that rotation function we don't need bc your model is radially symmetric
    morph.EAngleRotation = [0.0, 0.0, 0.0]

    # we will keep the validator in here. Always good to check the morphology your function made
    morph.validate_all(quiet=False)

    # this actually runs the model
    data = morph.run(stdout=True, stderr=False)

    # the PyHyperScattering part for interpreting the model result
    integrator = WPIntegrator(use_chunked_processing=False)
    remeshed_data = integrator.integrateImageStack(data)

    # we will use a pythin dictionary for holding the results
    results = {}

    results["I"] = remeshed_data.mean(dim="chi")
    results["A"] = remeshed_data.rsoxs.AR(chi_width=45)

    # note, in addition to the model output, you could also return other things like the visualization results (which you might want to see, for parameter sweeps)
    # morphology.visualize_materials has a huge number of potential arguments to tailor what it shows about the model and how it shows it
    results["visualizations"] = morph.visualize_materials(
        z_slice=args["vd"] // 2,
        subsample=256,
        outputplot=["vfrac", "S", "psi", "theta"],
        S_range=[[0, 1]],
        outputmat=[1],
        runquiet=True,
    )

    # also, it's a great idea to store the args dictionary in the results
    # that creates a durable record of what the parameters were when you simulated
    results["args"] = args

    # you could choose to return a variety of information.
    # for example, you could return the data object and/or the remeshed_data object
    # if you wanted to see full 2D patterns
    # I'm choosing only I and A, which are reductions of the data, but ones focused on our inquiry
    # I'm also returning the visualizations created above

    # note we return the results dictionary, which has I, A, and visualizations stored in it
    return results


# FROM NOTEBOOK 2
def adsorbed_polymer_run_save(args):
    """
    Saves the results of the adsorbed polymer run to a pickle file.

    Args:
        args (dict): A dictionary containing the arguments for running the model.

    Returns:
        None
    """
    # we jump straight to the results dictionary
    results = adsorbed_polymer_run(args)

    # now we save the results to a pickle file
    # note that we need a file name! We can use a new entry in the args dictionary
    # also we need  a directory to save the file in.
    # For convenience, I've hardcoded a "pickles" subdirectory in the current working directory

    # this syntax is using the pathlib module. Some older styles might just use strings for this
    # Path.cwd() returns the current working directory

    savefile_path_and_name = Path.cwd() / "pickles" / args["filename"]

    # lzma is a hardcore compression library we're using with pickle to keep the file size down
    with lzma.open(savefile_path_and_name, "wb") as handle:
        pickle.dump(results, handle, protocol=pickle.HIGHEST_PROTOCOL)

    # note there is nothing to return here. The results are stored in the pickled results dictionary

# FROM NOTEBOOK 3
def adsorbed_polymer_sweep(args, swept_arg, values, file_prefix = ""):
    """
    Runs the adsorbed polymer model for multiple values of a swept argument and saves the results to pickle files.

    Args:
        args (dict): A dictionary containing the arguments for running the model.
        swept_arg (str): The name of the argument to sweep.
        values (list): A list of values to sweep over.
        file_prefix (str, optional): A prefix to add to the filename. Defaults to an empty string.

    Returns:
        None
    """
        
    # note that the default value of the file_prefix is an empty string
    # this prefix is here just if you want to mark the filenames with a special marker like the date,
    # or other values you might have set outside this function like euler style

    # add a couple of values here to args including the swept args and values
    # this is mostly to save for after-simulation analysis
    args["swept_arg"] = swept_arg
    args["swept_values"] = values

    # enumerate provides two values, the index (0,1,2,3, etc) and the value.
    # we use that here so that the variable i can keep track of which value we're on
    # and we'll use that in the filename
    for i, value in enumerate(values):
        # set the swept argument in the args dictionary to the current value    
        args[swept_arg] = value
        
        # set the filename to the concatenation of the prefix, the swept argument, and the index
        # the index is zero padded to 2 digits so that the filename is always the same length
        args["filename"] = file_prefix + swept_arg + "_" + str(i).zfill(2) + ".pkl"
        
        #run the model and pickle the result
        adsorbed_polymer_run_save(args)
        
# FROM NOTEBOOK 5
def import_pickles(picklePath, sstring):
    # below is a "docstring" and it is used to describe the function
    """
    Imports pickle files from a specified directory that match a given string.

    Parameters:
        picklePath (str): The path to the directory containing the pickle files.
        sstring (str): The string that the file names must contain.

    Returns:
        list: A list of objects representing the loaded pickle data.
    """
    # initialize an empty list to store the results
    results = []
    # use glob to find all the files in the directory that match the string
    # note the wildcard "*" in the glob function
    file_list = glob.glob(f"{picklePath}/{sstring}*")
    # show us those filenames
    pprint(file_list)
    # sort them - this will put them in the order they were created
    file_list.sort()
    # loop through the files. Notice how tqdm works - it actually is a function around the list that is iterated through
    for file_name in tqdm(file_list, colour = "blue"):
        # open the file using lzma
        with lzma.open(file_name, "rb") as pickle_file:
            # load the pickle file and append it to the results list
            results.append(pickle.load(pickle_file))
    # return the whole results list
    return results

#FROM NOTEBOOK 5

def plot_result(result):
    """
    Plots the result of a polymer adsorption simulation using matplotlib.

    Parameters:
        result (dict): A dictionary containing the following keys:
            - "I" (numpy.ndarray): The mean of the remeshed data along the "chi" dimension.
            - "A" (numpy.ndarray): The result of the rsoxs.AR function applied to the remeshed data.
            - "visualizations" (dict): A dictionary containing visualizations of the simulation result.
            - "args" (dict): A copy of the arguments used to run the simulation.

    Returns:
        numpy.ndarray: A 3D RGB array representing the plot.

    Example usage:
        >>> result = adsorbed_polymer_run(args)
        >>> rgb_array = plot_result(result)
    """
    rc('font', size=14)

    dpi = 72
    h_size = 1920 / dpi
    v_size = 1080 / dpi

    fig = plt.subplots(figsize=(h_size, v_size), dpi=dpi)

    gs = gridspec.GridSpec(
        nrows=2,
        ncols=2,
        figure=fig,
        width_ratios=[1.0, 1.0],
        height_ratios=[1.0, 1.0],
        wspace=0.2,
        hspace=0.2,
    )

    ax1 = plt.subplot(gs[0:1, 0:1])
    ax2 = plt.subplot(gs[0:1, 1:2])
    ax3 = plt.subplot(gs[1:2, 0:1])
    ax4 = plt.subplot(gs[1:2, 1:2])

    qmax = 1
    
    result["I"].sel(energy = 284.7, method = 'nearest').plot(ax = ax1, yscale = "log", xlim = (0.05, 1.4))

    ax1.set_title(f"Intensity vs. q")
    ax1.set_ylim(top = 1E8, bottom = 1)
    ax1.set_xlim(left = 0, right = qmax)
    ax1.set_xlabel("q ($nm^{-1}$)")
    ax1.set_ylabel("I (a.u.) at 284.7 eV")
    
    annotation = f"radius_nm = " + "{:0.1f}".format(result['args']["radius_nm"]) + " nm\n"
    annotation += f"S0 = " + "{:0.3f}".format(result['args']["S0"]) + "\n"
    annotation += f"S_slope_per_nm = " + "{:0.3f}".format(result['args']["S_slope_per_nm"])
    ax1.annotate(annotation, xy = (0.7, 0.8), xycoords = "axes fraction")
    
    ax2.imshow(result["visualizations"][1])
    ax2.axis("off")

    ax3.axhline(y = 0, color = "black", linestyle = "dotted") 
    result["A"].sel(energy = 284.7, method = 'nearest').plot(ax = ax3, ylim = (-1, 1), color = "green", label = "284.7 eV")
    result["A"].sel(energy = 285.2, method = 'nearest').plot(ax = ax3, ylim = (-1, 1), color = "magenta", label = "285.2 eV")
    ax3.set_xlim(left = 0, right = qmax)
    ax3.set_title(f"Radial simulation result, A")
    ax3.set_xlabel("q ($nm^{-1}$)")
    ax3.set_ylabel("A")
    ax3.legend()

    result["A"].plot(xlim=(0,1.0), vmin = -1, vmax = 1, cmap = 'seismic', ax = ax4)
    ax4.set_xlim(left = 0, right = qmax)
    ax4.set_title(f"Radial simulation result, A")
    ax4.set_xlabel("q ($nm^{-1}$)")
    ax4.set_ylabel("energy (eV)")

    plt.subplots_adjust(left=0.05, bottom=0.10, right=0.95, top=0.95, wspace=0, hspace=0)
    fig[0].canvas.draw()
    width, height = fig[0].canvas.get_width_height()
    data = np.frombuffer(fig[0].canvas.tostring_rgb(), dtype=np.uint8)
    rgb_array = data.reshape(height, width, 3)
    return rgb_array

import subprocess, shlex

#FROM NOTEBOOK 5
def encode_video_loop(
    frames, output_file, frame_rate=3, still_frames=5, loops=1
):
    """
    Encode a video from a sequence of frames, with a specified number
    of loops and still frames in between each loop.

    Parameters
    ----------
    frames : list of numpy arrays
        Frames to be encoded into the video
    output_file : str
        Path to the output video file
    frame_rate : int, optional
        Frame rate of the output video. Defaults to 3.
    still_frames : int, optional
        Number of still frames to write in between each loop. Defaults to 5.
    loops : int, optional
        Number of loops to write. Defaults to 1.
    """
    width, height = frames[0].shape[1], frames[0].shape[0]
    cmd = f'ffmpeg -y -s {width}x{height} -pixel_format rgb24 -f rawvideo -r {frame_rate} -i pipe: -vcodec libx264 -pix_fmt yuv420p -crf 15 {output_file}'
    process = subprocess.Popen(shlex.split(cmd), stdin=subprocess.PIPE, stdout=subprocess.DEVNULL)
    
    for _ in range(loops):
        for frame in frames:
            process.stdin.write(frame.tobytes())
        for _ in range(still_frames):
            process.stdin.write(frames[-1].tobytes())
        for frame in reversed(frames):
            process.stdin.write(frame.tobytes())
        for _ in range(still_frames):
            process.stdin.write(frames[0].tobytes())
    
    process.stdin.close()
    process.wait()
    process.terminate()


import numpy as np
import warnings
import h5py
import datetime
from skimage import transform


def projected_sphere(X, Y, xc, yc, radius):
    '''

    Arguments
    ---------
    X,Y: np.ndarray (M,N)
        2D meshgrids of X and Y positions of real-space grid

    xc,yc: float
        floating point positions of the x and y center of the sphere

    radius: float
        radius of the sphere

    Returns
    -------
    P: np.ndarray(M,N)
        2D image of sphere density

    Example
    --------
    ```python
    NumXY = 2048
    PhysSize=5.0
    Radius=500

    x = y = np.linspace(-PhysSize*NumXY/2,PhysSize*NumXY/2,NumXY)
    X, Y = np.meshgrid(x,y,indexing='ij')
    Mat_1_unaligned = tyler.lib.util.projected_sphere(X,Y,0.0,0.0,Radius)
    Mat_2_unaligned = 1.0 - Mat_1_unaligned
    plt.imshow(Mat_1_unaligned)

    #reshape arrays for CyRSoXS
    Mat_1_unaligned = Mat_1_unaligned[np.newaxis,:,:]
    Mat_2_unaligned = Mat_2_unaligned[np.newaxis,:,:]
    ```


    '''
    Xc2 = np.square(X-xc)
    Yc2 = np.square(Y-yc)
    r2 = radius*radius
    mask = ((Xc2 + Yc2)<(r2))
    P = np.zeros_like(X)
    P[mask] = 2*np.sqrt(r2 - Xc2[mask] - Yc2[mask])
    P /= np.amax(P)
    return P


def center_sphere(NumXY, PhysSize, radius):
    if radius > PhysSize*NumXY/2:
        warnings.warn('Sphere diameter is greater than simulation box width')

    x = y = np.linspace(-PhysSize*NumXY/2, PhysSize*NumXY/2, NumXY)
    X, Y = np.meshgrid(x, y, indexing='ij')
    proj_sphere = projected_sphere(X, Y, 0, 0, radius)

    return proj_sphere


def center_sphere3D(NumXY, NumZ, PhysSize, radius, interpolation=False, scale=2):
    if (radius > PhysSize*NumXY/2) | (radius > PhysSize*NumZ/2):
        warnings.warn('Sphere diameter is greater than a simulation box dimension')
    if interpolation:
        x = y = z = np.linspace(-int(radius/PhysSize), int(radius/PhysSize), int(scale*2*radius/PhysSize))
        Z, Y, X = np.meshgrid(z, y, x,indexing='ij')
        Xc2 = np.square(X)
        Yc2 = np.square(Y)
        Zc2 = np.square(Z)
        r2 = (radius/PhysSize)**2
        mask = ((Xc2 + Yc2 + Zc2) <= (r2))
        P_subset = np.zeros_like(X)
        P_subset[mask] = 1
        P_subset = transform.downscale_local_mean(P_subset, (scale, scale, scale))
        P = np.zeros((NumZ, NumXY, NumXY))
        lowXY = int(P.shape[1]/2 - P_subset.shape[1]/2)
        highXY = int(P.shape[1]/2 + P_subset.shape[1]/2)
        lowZ = int(P.shape[0]/2 - P_subset.shape[0]/2)
        highZ = int(P.shape[0]/2 + P_subset.shape[0]/2)
        P[lowZ:highZ, lowXY:highXY, lowXY:highXY] = P_subset
    else:
        x = y = z = np.linspace(-int(radius/PhysSize), int(radius/PhysSize), int(2*radius/PhysSize))
        Z, Y, X = np.meshgrid(z, y, x, indexing='ij')
        Xc2 = np.square(X)
        Yc2 = np.square(Y)
        Zc2 = np.square(Z)
        r2 = (radius/PhysSize)**2
        mask = ((Xc2 + Yc2 + Zc2) <= (r2))
        P_subset = np.zeros_like(X)
        P_subset[mask] = 1
        P = np.zeros((NumZ, NumXY, NumXY))
        low = int(P.shape[0]/2 - P_subset.shape[0]/2)
        high = int(P.shape[0]/2 + P_subset.shape[0]/2)
        P[low:high, low:high, low:high] = P_subset
    return P


def write_sphere_hdf5(phi1, NumXY, PhysSize, radius, euler=True, author='PJD'):
    s1 = s2 = np.zeros((1, NumXY, NumXY, 3))
    phi1_out = phi1[np.newaxis, :, :]
    phi2_out = 1 - phi1_out

    if euler:
        # placeholder zeros since there is no alignment in this example
        psi1 = psi2 = np.zeros((1, NumXY, NumXY))
        theta1 = theta2 = psi1.copy()
        S1 = S2 = psi1.copy()
    phys_str = str(PhysSize)
    phys_str = phys_str.replace('.', 'p')
    radius_str = str(radius)
    radius_str = radius_str.replace('.', 'p')
    fname = f'SingleSphere-{NumXY}-{phys_str}-{radius_str}.hdf5'
    print(f'--> Marking {fname}')

    array_ordering = 'ZYX'
    with h5py.File(fname, 'w') as f:
        if euler:
            f.create_dataset("Euler_Angles/Mat_1_Psi", data=psi1, compression='gzip', compression_opts=9)
            f.create_dataset("Euler_Angles/Mat_1_Theta", data=theta1, compression='gzip', compression_opts=9)
            f.create_dataset("Euler_Angles/Mat_1_S", data=S1, compression='gzip', compression_opts=9)
            f.create_dataset("Euler_Angles/Mat_1_Vfrac", data=phi1_out, compression='gzip', compression_opts=9)

            f.create_dataset("Euler_Angles/Mat_2_Psi", data=psi2, compression='gzip', compression_opts=9)
            f.create_dataset("Euler_Angles/Mat_2_Theta", data=theta2, compression='gzip', compression_opts=9)
            f.create_dataset("Euler_Angles/Mat_2_S", data=S2, compression='gzip', compression_opts=9)
            f.create_dataset("Euler_Angles/Mat_2_Vfrac", data=phi2_out, compression='gzip', compression_opts=9)
            for key in f['Euler_Angles'].keys():
                for i in range(3):
                    f['Euler_Angles'][key].dims[i].label = array_ordering[i]

        else:
            f.create_dataset("vector_morphology/Mat_1_alignment", data=s1, compression='gzip', compression_opts=9)
            f.create_dataset("vector_morphology/Mat_2_alignment", data=s2, compression='gzip', compression_opts=9)
            f.create_dataset("vector_morphology/Mat_1_unaligned", data=phi1_out, compression='gzip', compression_opts=9)
            f.create_dataset("vector_morphology/Mat_2_unaligned", data=phi2_out, compression='gzip', compression_opts=9)
            for key in f['vector_morphology'].keys():
                for i in range(3):
                    f['vector_morphology'][key].dims[i].label = array_ordering[i]

        f.create_dataset('Morphology_Parameters/NumMaterial', data=2)
        f.create_dataset('Morphology_Parameters/PhysSize', data=PhysSize)
        f.create_dataset('Morphology_Parameters/creation_date', data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
        f.create_dataset('Morphology_Parameters/film_normal', data=[1, 0, 0])
        f.create_dataset('Morphology_Parameters/morphology_creator', data=author)
        f.create_dataset('Morphology_Parameters/name', data=author)
        f.create_dataset('Morphology_Parameters/version', data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
#             f.create_dataset('Morphology_Parameters/voxel_size_nm', data=PhysSize)

        f.create_dataset('igor_parameters/igorefield', data="0,1")
        f.create_dataset('igor_parameters/igormaterials', data="PEOlig2018,vac")
        f.create_dataset('igor_parameters/igormodelname', data="SingleSphere")
        f.create_dataset('igor_parameters/igormovie', data=0)
        f.create_dataset('igor_parameters/igorname', data="perp001")
        f.create_dataset('igor_parameters/igornum', data=0)
        f.create_dataset('igor_parameters/igorparamstring', data="n/a")
        f.create_dataset('igor_parameters/igorpath', data="n/a")
        f.create_dataset('igor_parameters/igorrotation', data=0)
        f.create_dataset('igor_parameters/igorthickness', data=1)
        f.create_dataset('igor_parameters/igorvoxelsize', data=1)
    return fname


def write_config2D(NumXY, PhysSize,
                   energies, euler=False, referenceFrame=1):
    f = open("config.txt", "w")
    f.write('Energies = ' + str(list(energies)) + ';\n')
    f.write('CaseType=0;  # leave it as it is \n')
    f.write('EAngleRotation=[0.0,1.0,360.0]; # Angle Rotation \n')

    f.write('MorphologyType=' + str(1-int(euler)) + ';  # 0: Euler 1: Vector\n')
    f.write('Algorithm=0;  # Leave it as it is\n')
    f.write("NumThreads = 4;\n")
    f.write("NumX = " + str(NumXY) + ";\n")
    f.write("NumY = " + str(NumXY) + ";\n")
    f.write("NumZ = " + str(1) + ";\n")
    f.write("PhysSize = " + str(PhysSize) + ";\n")
    f.write("ReferenceFrame = " + str(referenceFrame) + ";\n")
    f.close()


def write_slurm(filename, cyrsoxs_version='latest'):

    f = open("job.slurm","w")
    f.write('#!/bin/bash -l\n')
    f.write('#SBATCH --job-name=CYRSOXS      # Job name\n')
    f.write('#SBATCH --output=cyrsoxs.%j.out # Stdout (%j expands to jobId)\n')
    f.write('#SBATCH --error=cyrsoxs.%j.out  # Stderr (%j expands to jobId)\n')
    f.write('#SBATCH --time=04:00:00         # walltime\n')
    f.write('#SBATCH --nodes=1               # Number of nodes requested (only one node on glados)\n')
    f.write('#SBATCH --ntasks=1              # Number of tasks(processes) (tasks distributed across nodes)\n')
    f.write('#SBATCH --ntasks-per-node=1     # Tasks per node\n')
    f.write('#SBATCH --cpus-per-task=1       # Threads per task (all cpus will be on same node)\n')
    f.write('#SBATCH --gres=gpu:turing:1     # number and type of GPUS to use (glados only has turing)\n')
    f.write('#SBATCH --gres-flags=enforce-binding\n')
    f.write('#SBATCH --partition=gpu\n')
    f.write('\n')
    f.write('set -e\n')
    f.write('\n')
    f.write('if [ x$SLURM_CPUS_PER_TASK == x ]; then\n')
    f.write('    export OMP_NUM_THREADS=1\n')
    f.write('else\n')
    f.write('    export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n')
    f.write('fi\n')
    f.write('\n')
    f.write('# load necessary modules\n')
    f.write('source ~/.typyEnv CLEAN\n')
    f.write(f'typyEnv --add cyrsoxs/{cyrsoxs_version}\n')
    f.write('\n')
    f.write('## RUN YOUR PROGRAM ##\n')
    f.write('echo "RUNNING ON GPU"${CUDA_VISIBLE_DEVICES}\n')
    f.write(f'srun CyRSoXS {filename}')
    f.close()


def write_sphere3D_hdf5(phi1, NumXYZ, PhysSize, radius, author='PJD'):
    s1 = s2 = np.zeros((NumXYZ, NumXYZ, NumXYZ, 3))
    phi1_out = phi1[np.newaxis, :, :]
    phi2_out = 1 - phi1_out

    fname = f'SingleSphere-{NumXYZ}-{PhysSize}-{radius}.hdf5'
    print(f'--> Marking {fname}')
    with h5py.File(fname, 'w') as f:
        f.create_dataset("igor_parameters/igormaterialnum", data=2.0)
        f.create_dataset("vector_morphology/Mat_1_alignment", data=s1, compression='gzip', compression_opts=9)
        f.create_dataset("vector_morphology/Mat_2_alignment", data=s2, compression='gzip', compression_opts=9)
        f.create_dataset("vector_morphology/Mat_1_unaligned", data=phi1_out, compression='gzip', compression_opts=9)
        f.create_dataset("vector_morphology/Mat_2_unaligned", data=phi2_out, compression='gzip', compression_opts=9)

        f.create_dataset('morphology_variables/creation_date', data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
        f.create_dataset('morphology_variables/film_normal', data=[1, 0, 0])
        f.create_dataset('morphology_variables/morphology_creator', data=author)
        f.create_dataset('morphology_variables/name', data=author)
        f.create_dataset('morphology_variables/version', data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
        f.create_dataset('morphology_variables/voxel_size_nm', data=PhysSize)

        f.create_dataset('igor_parameters/igorefield', data="0,1")
        f.create_dataset('igor_parameters/igormaterials', data="PEOlig2018,vac")
        f.create_dataset('igor_parameters/igormodelname', data="SingleSphere")
        f.create_dataset('igor_parameters/igormovie', data=0)
        f.create_dataset('igor_parameters/igorname', data="perp001")
        f.create_dataset('igor_parameters/igornum', data=0)
        f.create_dataset('igor_parameters/igorparamstring', data="n/a")
        f.create_dataset('igor_parameters/igorpath', data="n/a")
        f.create_dataset('igor_parameters/igorrotation', data=0)
        f.create_dataset('igor_parameters/igorthickness', data=1)
        f.create_dataset('igor_parameters/igorvoxelsize', data=1)
    return fname


def write_config3D(phi1, PhysSize, numXYZ,
                   startEnergy, endEnergy, incrementEnergy,
                   startAngle=0.0, endAngle=360.0, incrementAngle=2.0,
                   numThreads=4):
    f = open("config.txt", "w")
    f.write("StartEnergy = " + str(startEnergy) + ";\n")
    f.write("EndEnergy = " + str(endEnergy) + ";\n")
    f.write("IncrementEnergy = " + str(incrementEnergy) + ";\n")
    f.write("StartAngle = " + str(startAngle) + ";\n")
    f.write("EndAngle = " + str(endAngle) + ";\n")
    f.write("IncrementAngle = " + str(incrementAngle) + ";\n")
    f.write("NumThreads = " + str(numThreads) + ";\n")
    f.write("NumX = " + str(numXYZ) + ";\n")
    f.write("NumY = " + str(numXYZ) + ";\n")
    f.write("NumZ = " + str(numXYZ) + ";\n")
    f.write("PhysSize = " + str(PhysSize) + ";\n")
    f.close()


def write_grating_hdf5(s1, s2, phi1_out, phi2_out, PhysSize, width, author='PJD'):
    NumX = phi1_out.shape[1]
    NumY = phi1_out.shape[2]
    NumZ = phi1_out.shape[0]

    fname = f'SingleGrating-{NumX}-{NumY}-{NumZ}-{PhysSize}-{width}.hdf5'
    print(f'--> Marking {fname}')
    with h5py.File(fname, 'w') as f:
        f.create_dataset("igor_parameters/igormaterialnum", data=2.0)
        f.create_dataset("vector_morphology/Mat_1_alignment", data=s1, compression='gzip', compression_opts=9)
        f.create_dataset("vector_morphology/Mat_2_alignment", data=s2, compression='gzip', compression_opts=9)
        f.create_dataset("vector_morphology/Mat_1_unaligned", data=phi1_out, compression='gzip', compression_opts=9)
        f.create_dataset("vector_morphology/Mat_2_unaligned", data=phi2_out, compression='gzip', compression_opts=9)

        f.create_dataset('morphology_variables/creation_date', data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
        f.create_dataset('morphology_variables/film_normal', data=[1, 0, 0])
        f.create_dataset('morphology_variables/morphology_creator', data=author)
        f.create_dataset('morphology_variables/name', data=author)
        f.create_dataset('morphology_variables/version', data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
        f.create_dataset('morphology_variables/voxel_size_nm', data=PhysSize)

        f.create_dataset('igor_parameters/igorefield', data="0,1")
        f.create_dataset('igor_parameters/igormaterials', data="PEOlig2018,vac")
        f.create_dataset('igor_parameters/igormodelname', data="SingleGrating")
        f.create_dataset('igor_parameters/igormovie', data=0)
        f.create_dataset('igor_parameters/igorname', data="perp001")
        f.create_dataset('igor_parameters/igornum', data=0)
        f.create_dataset('igor_parameters/igorparamstring', data="n/a")
        f.create_dataset('igor_parameters/igorpath', data="n/a")
        f.create_dataset('igor_parameters/igorrotation', data=0)
        f.create_dataset('igor_parameters/igorthickness', data=1)
        f.create_dataset('igor_parameters/igorvoxelsize', data=1)
    return fname
